# flake8: noqa
from __future__ import annotations

from typing import Optional

from keplemon.elements import CartesianState, GeodeticPosition, KeplerianElements, KeplerianState, TLE
from keplemon.time import Epoch

class ForceProperties:
    srp_coefficient: float
    drag_coefficient: float
    mass: float
    srp_area: float
    drag_area: float
    mean_motion_dot: float
    mean_motion_dot_dot: float
    def __init__(
        self,
        srp_coefficient: float,
        srp_area: float,
        drag_coefficient: float,
        drag_area: float,
        mass: float,
        mean_motion_dot: float,
        mean_motion_dot_dot: float,
    ) -> None: ...

class InertialPropagator:
    @staticmethod
    def from_tle(tle: TLE) -> InertialPropagator: ...

    def get_cartesian_state_at_epoch(self, epoch: Epoch) -> Optional[CartesianState]: ...
    def get_keplerian_state_at_epoch(self, epoch: Epoch) -> Optional[KeplerianState]: ...

    keplerian_state: KeplerianState
    force_properties: ForceProperties

class SGP4Output:
    cartesian_state: CartesianState
    mean_elements: KeplerianElements
    osculating_elements: KeplerianElements
    geodetic_position: GeodeticPosition

def b_star_to_drag_coefficient(b_star: float) -> float:
    """Convert B* to drag coefficient."""
    ...

def drag_coefficient_to_b_star(drag_coefficient: float) -> float:
    """Convert drag coefficient to B*."""
    ...



# flake8: noqa
from keplemon.time import Epoch, TimeSpan
from keplemon.elements import HorizonState, CartesianVector, TopocentricElements
from keplemon.enums import ReferenceFrame, UCTObservability, UCTValidity
from keplemon.estimation import ObservationAssociation

class FieldOfViewCandidate:
    satellite_id: str
    """ID of the candidate satellite"""

    direction: TopocentricElements
    """Measured direction to the candidate satellite in the sensor's topocentric frame"""

class FieldOfViewReport:
    epoch: Epoch
    """UTC epoch of the field of view report"""

    sensor_position: CartesianVector
    """TEME position of the sensor in the observatory's topocentric frame in **_kilometers_**"""

    sensor_direction: TopocentricElements
    """Direction of the sensor in the observatory's topocentric frame"""

    fov_angle: float
    """Field of view angle of the sensor in **_degrees_**"""

    candidates: list[FieldOfViewCandidate]
    """List of candidate satellites within the field of view"""

    reference_frame: ReferenceFrame
    """Reference frame of the output direction elements"""

class CloseApproach:
    epoch: Epoch
    """UTC epoch of the close approach"""

    primary_id: str
    """Satellite ID of the primary body in the close approach"""

    secondary_id: str
    """Satellite ID of the secondary body in the close approach"""

    distance: float
    """Distance between the two bodies in **_kilometers_**"""

class CloseApproachReport:
    """
    Args:
        start: CA screening start time
        end: CA screening end time
        distance_threshold: Distance threshold for CA screening in **_kilometers_**
    """

    close_approaches: list[CloseApproach]
    """List of close approaches found during the screening"""

    distance_threshold: float
    def __init__(self, start: Epoch, end: Epoch, distance_threshold: float) -> None: ...

class HorizonAccess:

    satellite_id: str
    """ID of the satellite for which the access is calculated"""

    observatory_id: str
    """ID of the observatory for which the access is calculated"""

    start: HorizonState
    """State of the satellite at the start of the access period"""

    end: HorizonState
    """State of the satellite at the end of the access period"""

class HorizonAccessReport:
    """
    Args:
        start: UTC epoch of the start of the access report
        end: UTC epoch of the end of the access report
        min_elevation: Minimum elevation angle for access in **_degrees_**
        min_duration: Minimum duration of access
    """

    accesses: list[HorizonAccess]
    """List of horizon accesses found during the screening"""

    elevation_threshold: float
    """Minimum elevation angle for access in **_degrees_**"""

    start: Epoch
    """UTC epoch of the start of the access report"""

    end: Epoch
    """UTC epoch of the end of the access report"""

    duration_threshold: TimeSpan
    """Minimum duration of a valid access"""

    def __init__(
        self,
        start: Epoch,
        end: Epoch,
        min_elevation: float,
        min_duration: TimeSpan,
    ) -> None: ...

class ProximityEvent:
    """Represents a time period where two satellites remain within a distance threshold."""

    primary_id: str
    """Satellite ID of the primary body"""

    secondary_id: str
    """Satellite ID of the secondary body"""

    start_epoch: Epoch
    """UTC epoch of the start of the proximity event"""

    end_epoch: Epoch
    """UTC epoch of the end of the proximity event"""

    minimum_distance: float
    """Minimum distance between the two bodies during the event in **_kilometers_**"""

    maximum_distance: float
    """Maximum distance between the two bodies during the event in **_kilometers_**"""

class ProximityReport:
    """
    Args:
        start: Proximity screening start time
        end: Proximity screening end time
        distance_threshold: Distance threshold for proximity screening in **_kilometers_**
    """

    events: list[ProximityEvent]
    """List of proximity events found during the screening"""

    distance_threshold: float
    """Distance threshold for proximity screening in **_kilometers_**"""

    start: Epoch
    """UTC epoch of the start of the proximity report"""

    end: Epoch
    """UTC epoch of the end of the proximity report"""

    def __init__(self, start: Epoch, end: Epoch, distance_threshold: float) -> None: ...

class ManeuverEvent:
    """Represents a detected maneuver for a satellite."""

    satellite_id: str
    """Satellite ID of the maneuvering body"""

    epoch: Epoch
    """UTC epoch of the detected maneuver"""

    delta_v: CartesianVector
    """Delta-V vector in RIC (radial, in-track, cross-track) frame in **_meters per second_**"""

class ManeuverReport:
    """
    Args:
        start: Maneuver detection start time
        end: Maneuver detection end time
        distance_threshold: Distance threshold for matching in **_kilometers_**
        velocity_threshold: Velocity threshold for maneuver detection in **_meters per second_**
    """

    maneuvers: list[ManeuverEvent]
    """List of detected maneuvers"""

    distance_threshold: float
    """Distance threshold for matching in **_kilometers_**"""

    velocity_threshold: float
    """Velocity threshold for maneuver detection in **_meters per second_**"""

    start: Epoch
    """UTC epoch of the start of the maneuver report"""

    end: Epoch
    """UTC epoch of the end of the maneuver report"""

    def __init__(
        self, start: Epoch, end: Epoch, distance_threshold: float, velocity_threshold: float
    ) -> None: ...

class UCTValidityReport:
    """
    Report containing UCT validity analysis results.

    This report provides information about a UCT's validity based on:
    - Observation associations with the UCT
    - Proximity events with approved satellites (possible cross-tags)
    - Close approaches with approved satellites (possible maneuver origins)
    - Observability status during the analysis window
    """

    satellite_id: str
    """ID of the UCT satellite being analyzed"""

    associations: list[ObservationAssociation]
    """List of observation associations where the UCT matched orphan observations"""

    possible_cross_tags: list[ProximityEvent]
    """List of proximity events with approved satellites that could indicate cross-tagging"""

    possible_origins: list[CloseApproach]
    """List of close approaches with approved satellites that could indicate maneuver origins"""

    observability: UCTObservability
    """Observability status of the UCT during the analysis window"""

    validity: UCTValidity
    """Validity assessment based on the analysis results"""



# flake8: noqa
from enum import Enum

class AssociationConfidence(Enum):
    """
    Confidence level of an observation association

    Attributes:
        Low (AssociationConfidence): Low confidence
        Medium (AssociationConfidence): Medium confidence
        High (AssociationConfidence): High confidence
    """

    Low = ...
    Medium = ...
    High = ...

class CovarianceType(Enum):
    """
    Reference frame/element types for a covariance matrix

    Attributes:
        Inertial (CovarianceType): Cartesian TEME
        Relative (CovarianceType): Cartesian UVW
        Equinoctial (CovarianceType): Equinoctial
    """

    Inertial = ...
    Relative = ...
    Equinoctial = ...

class GeodeticModel(Enum):
    """
    Geodetic model used for Earth shape and gravity

    Attributes:
        WGS72 (GeodeticModel): WGS72 geodetic model
        WGS84 (GeodeticModel): WGS84 geodetic model
        EGM96 (GeodeticModel): EGM96 geodetic model
    """

    WGS72 = ...
    WGS84 = ...
    EGM96 = ...

class MeanEquinox(Enum):
    """
    Mean equinox used for topocentric calculations

    Attributes:
        OfDate (MeanEquinox): Mean equinox of date
        OfYear (MeanEquinox): Mean equinox of year
        J2000 (MeanEquinox): J2000 mean equinox
        B1950 (MeanEquinox): B1950 mean equinox
    """

    OfDate = ...
    OfYear = ...
    J2000 = ...
    B1950 = ...
    value: int

    def get_value(self) -> int: ...

class TimeSystem(Enum):
    """
    Attributes:
        UTC (TimeSystem): Coordinated Universal Time
        TAI (TimeSystem): International Atomic Time
        TT (TimeSystem): Terrestrial Time
        UT1 (TimeSystem): Universal Time
    """

    UTC = ...
    TAI = ...
    TT = ...
    UT1 = ...

class Classification(Enum):
    """
    Simple classification primarily used to construct single-character identifiers in SAAL data

    Attributes:
        Unclassified (Classification): Unclassified
        Confidential (Classification): Confidential
        Secret (Classification): Secret
    """

    Unclassified = ...
    Confidential = ...
    Secret = ...

class KeplerianType(Enum):
    """
    Theory used to construct the Keplerian elements

    Attributes:
        MeanKozaiGP (KeplerianType): SGP4 mean elements with Kozai mean motion
        MeanBrouwerGP (KeplerianType): SGP4 mean elements with Brouwer mean motion
        MeanBrouwerXP (KeplerianType): SGP4-XP mean elements with Brouwer mean motion
        Osculating (KeplerianType): Osculating elements with Brouwer mean motion
    """

    MeanKozaiGP = ...
    MeanBrouwerGP = ...
    MeanBrouwerXP = ...
    Osculating = ...

class ReferenceFrame(Enum):
    """
    Reference frame used for inertial elements

    Attributes:
        TEME (ReferenceFrame): True Equator Mean Equinox
        J2000 (ReferenceFrame): J2000
        EFG (ReferenceFrame): Earth Fixed Greenwich (no polar motion)
        ECR (ReferenceFrame): Earth Centered Rotating (polar motion)
    """

    TEME = ...
    J2000 = ...
    EFG = ...
    ECR = ...

class UCTObservability(Enum):
    """
    Observability status of a UCT during analysis.

    Attributes:
        Confirmed (UCTObservability): UCT was confirmed to be observable (matched orphan observations)
        Possible (UCTObservability): UCT was within sensor field of view but not confirmed
        Unavailable (UCTObservability): UCT was not observable during the analysis window
    """

    Confirmed = ...
    Possible = ...
    Unavailable = ...
    value: str

class UCTValidity(Enum):
    """
    Validity assessment of a UCT based on analysis.

    Attributes:
        LikelyReal (UCTValidity): UCT is likely a real distinct object (has confirmed observations)
        PossibleCrossTag (UCTValidity): UCT may be a cross-tag of an approved satellite (proximity events found)
        PossibleManeuver (UCTValidity): UCT may be result of a maneuver (close approaches found)
        Inconclusive (UCTValidity): Insufficient evidence to determine UCT validity
    """

    LikelyReal = ...
    PossibleCrossTag = ...
    PossibleManeuver = ...
    Inconclusive = ...
    value: str



from keplemon._keplemon.elements import (  # type: ignore
    TLE,
    KeplerianState,
    CartesianState,
    CartesianVector,
    KeplerianElements,
    Ephemeris,
    SphericalVector,
    TopocentricElements,
    HorizonState,
    HorizonElements,
    GeodeticPosition,
    OrbitPlotData,
    OrbitPlotState,
    RelativeState,
    BoreToBodyAngles,
    TopocentricState,
    EquinoctialElements,
)

__all__ = [
    "TLE",
    "KeplerianState",
    "CartesianState",
    "CartesianVector",
    "KeplerianElements",
    "Ephemeris",
    "SphericalVector",
    "TopocentricElements",
    "HorizonState",
    "HorizonElements",
    "GeodeticPosition",
    "OrbitPlotData",
    "OrbitPlotState",
    "RelativeState",
    "BoreToBodyAngles",
    "TopocentricState",
    "EquinoctialElements",
]



from keplemon._keplemon.time import (  # type: ignore
    TimeSpan,
    TimeComponents,
    Epoch,
)
import requests  # type: ignore
from datetime import datetime
from keplemon._keplemon.enums import TimeSystem  # type: ignore

__all__ = [
    "TimeSpan",
    "TimeComponents",
    "Epoch",
    "request_time_constants_update",
]


def request_time_constants_update(output_path: str) -> None:
    finals = requests.get("https://maia.usno.navy.mil/ser7/finals.all").text.splitlines()

    leap_seconds = requests.get("https://maia.usno.navy.mil/ser7/tai-utc.dat").text.splitlines()

    month_map = {
        "JAN": 1,
        "FEB": 2,
        "MAR": 3,
        "APR": 4,
        "MAY": 5,
        "JUN": 6,
        "JUL": 7,
        "AUG": 8,
        "SEP": 9,
        "OCT": 10,
        "NOV": 11,
        "DEC": 12,
    }
    leap_second_map = {}
    for line in leap_seconds:
        yy = int(line[3:5].strip())
        mm = month_map[line[6:9]]
        dd = int(line[10:12].strip())
        time_comps = TimeComponents(yy, mm, dd, 0, 0, 0)
        current_epoch = Epoch.from_time_components(time_comps, TimeSystem.UTC)
        ds50 = int(current_epoch.days_since_1950)
        leap = float(line[38:48].strip())
        leap_second_map[ds50] = leap

    time_constant_lines: list[str] = []
    sorted_leap_seconds = sorted(leap_second_map, reverse=True)
    ut1 = 0.0
    prev_ut1_utc = 0.0
    prev_time = 0
    prev_leap = 0.0
    for line in finals:

        if len(line.strip()) < 68:
            break

        yy = int(line[0:2].strip())
        mm = int(line[2:4].strip())
        dd = int(line[4:6].strip())
        if yy < 50:
            yyyy = 2000 + yy
        else:
            yyyy = 1900 + yy
        ymd = datetime(yyyy, mm, dd).strftime("%d-%b-%y")

        time_comps = TimeComponents(yy, mm, dd, 0, 0, 0)
        current_epoch = Epoch.from_time_components(time_comps, TimeSystem.UTC)
        ds50 = int(current_epoch.days_since_1950)
        leap = 0.0
        for leap_ds50 in sorted_leap_seconds:
            if ds50 >= leap_ds50:
                leap = leap_second_map[leap_ds50]
                break

        ut1 = float(line[58:68].strip())
        if prev_time == 0:
            ut1_r = 0.0
        else:
            ut1_r = (ut1 - prev_ut1_utc) / (ds50 - prev_time)
            ut1_r = (ut1_r - (leap - prev_leap)) * 1e3

        p_x = float(line[18:27].strip())
        p_y = float(line[37:46].strip())

        doy = current_epoch.day_of_year
        time_constant_lines.append(
            f" {yy:02d}  {doy:03.0f}   {ymd} {leap:3.0f}  {ut1:8.5f}  {ut1_r:6.3f}   {p_x:7.4f}   {p_y:7.4f}\n"
        )
        prev_time = ds50
        prev_ut1_utc = ut1
        prev_leap = leap

    with open(output_path, "w") as f:
        f.writelines(time_constant_lines)



# flake8: noqa
from typing import Optional

from keplemon.elements import TopocentricElements, CartesianVector
from keplemon.time import Epoch
from keplemon.bodies import Satellite, Sensor, Constellation
from keplemon.enums import KeplerianType, AssociationConfidence

class ObservationCollection:
    """
    A collection of simultaneous observations from the same sensor position.

    Args:
        obs: List of observations with the same epoch and observer position
    """

    sensor_position: CartesianVector
    """TEME position of the sensor in **_kilometers_**"""

    sensor_direction: CartesianVector
    """Unit vector pointing in the average direction of all observations"""

    field_of_view: float
    """Estimated field of view in **_degrees_** based on observation spread"""

    observations: list["Observation"]
    """List of observations in this collection"""

    epoch: "Epoch"
    """Common epoch of all observations in the collection"""

    def __init__(self, obs: list["Observation"]) -> None: ...
    @staticmethod
    def get_list(obs: list["Observation"]) -> list["ObservationCollection"]:
        """
        Group observations by epoch and observer position into collections.

        Args:
            obs: List of observations with potentially different epochs and positions

        Returns:
            List of ObservationCollection instances, one for each unique (epoch, position) pair
        """
        ...

    def get_visibility(self, satellite: "Satellite") -> bool:
        """
        Check if a satellite is within the field of view of this collection.

        Args:
            satellite: Satellite to check visibility for

        Returns:
            True if the satellite is within the field of view
        """
        ...

    def get_association(self, satellite: "Satellite") -> Optional["ObservationAssociation"]:
        """
        Get the best observation association for a satellite within this collection.

        Args:
            satellite: Satellite to find an association for

        Returns:
            Best matching ObservationAssociation, or None if not visible or no match
        """
        ...

    def get_association_report(self, satellites: "Constellation") -> "CollectionAssociationReport":
        """
        Get optimal observation-satellite associations for a constellation.

        This method finds the best combination of observation-satellite pairs,
        prioritizing high confidence associations, then medium, then low.
        Each satellite can only associate with one observation and vice versa.

        Args:
            satellites: Constellation of satellites to find associations for

        Returns:
            CollectionAssociationReport containing associations, orphans, and moving satellite IDs
        """
        ...

class Covariance:
    sigmas: list[float]
    """"""

class CollectionAssociationReport:
    """Report of observation associations for an ObservationCollection."""

    orphan_observations: list["Observation"]
    """Observations that could not be associated with any satellite"""

    associations: list["ObservationAssociation"]
    """List of observation-satellite associations found"""

    moving_satellite_ids: list[str]
    """Satellite IDs with low or medium confidence associations (potential movers)"""

class ObservationAssociation:
    observation_id: str
    satellite_id: str
    residual: "ObservationResidual"
    confidence: "AssociationConfidence"

class Observation:
    """
    Args:
        sensor: Sensor that made the observation
        epoch: Time of the observation
        observed_teme_topocentric: Topocentric elements of the satellite at the time of observation
        observer_teme_position: Position of the observer in TEME coordinates
    """

    id: str
    """Unique identifier for the observation"""

    sensor: Sensor
    """Sensor which produced the observation"""

    epoch: Epoch
    """Time the measurement was observed"""

    range: Optional[float]
    """Observed range from the sensor to the satellite in **_kilometers_**"""

    range_rate: Optional[float]
    """Observed range rate from the sensor to the satellite in **_kilometers per second_**"""

    right_ascension: float
    """Observed TEME right ascension in **_degrees_**"""

    declination: float
    """Observed TEME declination in **_degrees_**"""

    right_ascension_rate: Optional[float]
    """Observed right ascension rate in **_degrees per second_**"""

    declination_rate: Optional[float]
    """Observed declination rate in **_degrees per second_**"""

    observed_satellite_id: Optional[str]
    """Tagged satellite ID of the observation"""

    def __init__(
        self,
        sensor: Sensor,
        epoch: Epoch,
        observed_teme_topocentric: TopocentricElements,
        observer_teme_position: CartesianVector,
    ) -> None: ...
    @staticmethod
    def from_saal_files(sensor_file: str, observation_file: str) -> list["Observation"]: ...
    def get_residual(self, sat: Satellite) -> Optional[ObservationResidual]:
        """
        Calculate the residual of the observation with respect to a given satellite state.

        !!! note
            If an error occurs during propagation of the satellite state, this method will return None.

        Args:
            sat: Expected satellite state

        Returns:
            Calculated residual
        """
        ...

    def get_associations(self, sats: Constellation) -> list[ObservationAssociation]:
        """
        Calculate the associations of the observation with respect to a given constellation of satellites.

        Args:
            sats: Constellation of satellites to compare against

        Returns:
            List of possible observation associations
        """
        ...

class ObservationResidual:
    range: float
    """Euclidean distance between the observed and expected state in **_kilometers_**"""

    radial: float
    """Radial distance between the observed and expected state in **_kilometers_**"""

    in_track: float
    """In-track distance between the observed and expected state in **_kilometers_**"""

    cross_track: float
    """Cross-track distance between the observed and expected state in **_kilometers_**"""

    velocity: float
    """Velocity magnitude difference between the observed and expected state in **_kilometers per second_**"""

    radial_velocity: float
    """Radial velocity difference between the observed and expected state in **_kilometers per second_**"""

    in_track_velocity: float
    """In-track velocity difference between the observed and expected state in **_kilometers per second_**"""

    cross_track_velocity: float
    """Cross-track velocity difference between the observed and expected state in **_kilometers per second_**"""

    time: float
    """Time difference between the observed and expected state in **_seconds_**"""

    beta: float
    """Out-of-plane difference between the observed and expected state in **_degrees_**"""

    height: float
    """Height difference between the observed and expected state in **_kilometers_**"""

class BatchLeastSquares:
    """
    Args:
        obs: List of observations to be used in the estimation
        a_priori: A priori satellite state
    """

    converged: bool
    """Indicates if the solution meets the tolerance criteria"""

    max_iterations: int
    """Maximum number of iterations to perform when solving if the tolerance is not met"""

    iteration_count: int
    """Number of iterations performed to reach the solution"""

    current_estimate: Satellite
    """Current estimate of the satellite state after iterating or solving"""

    rms: Optional[float]
    """Root mean square of the residuals in **_kilometers_**"""

    weighted_rms: Optional[float]
    """Unitless weighted root mean square of the residuals"""

    estimate_srp: bool
    """Flag to indicate if solar radiation pressure should be estimated
    
    !!! warning
        This currently has unexpected behavior if solving for output_types other than XP
    """

    estimate_drag: bool
    """Flag to indicate if atmospheric drag should be estimated
    
    !!! warning
        This currently has unexpected behavior if solving for output_types other than XP
    """

    a_priori: Satellite
    """A priori satellite state used to initialize the estimation"""

    observations: list[Observation]
    """List of observations used in the estimation"""

    residuals: list[tuple[Epoch, ObservationResidual]]
    """List of residuals for each observation compared to the current estimate"""

    covariance: Optional[Covariance]
    """UVW covariance matrix of the current estimate in **_kilometers_** and **_kilometers per second_**"""

    output_type: KeplerianType
    """Type of Keplerian elements to be used in the output state"""

    eccentricity_constraint_weight: Optional[float]
    """Tikhonov weight that keeps equinoctial a_f/a_g (eccentricity) near the a priori state"""

    estimate_maneuver: bool
    """Enable maneuver estimation mode. When True, disables drag/SRP estimation and
    searches for a maneuver epoch in the last half orbital period."""

    maneuver_epoch: Optional[Epoch]
    """Detected maneuver epoch after solve() if estimate_maneuver was True.
    None if no significant maneuver was detected (delta-V < 0.05 m/s)."""

    delta_v: Optional[CartesianVector]
    """Delta-V in RIC frame (radial, in-track, cross-track) in **_km/s_** after solve()
    if estimate_maneuver was True. None if no significant maneuver was detected."""

    allow_radial_delta_v: bool
    """When False, heavily penalizes radial delta-V component during maneuver epoch search,
    forcing the solution toward pure in-track maneuvers. Default True (no penalty)."""

    def __init__(
        self,
        obs: list[Observation],
        a_priori: Satellite,
    ) -> None: ...
    def solve(self) -> None:
        """Iterate until the solution converges or the maximum number of iterations is reached."""
        ...

    def iterate(self) -> None:
        """Perform a single iteration of the estimation process."""
        ...

    def reset(self) -> None:
        """Reset the estimation process to the initial state."""
        ...



# flake8: noqa
from __future__ import annotations
from pathlib import Path
from keplemon.enums import TimeSystem
from typing import overload, Any, Union
from datetime import datetime

def request_time_constants_update(output_path: Union[str, Path]) -> None:
    """
    Request time constants and EOP data from USNO and save to the specified path.

    Args:
        output_path: Path where the SAAL-formatted time constants will be written
    """
    ...

class TimeSpan:
    @classmethod
    def from_days(cls, days: float) -> TimeSpan:
        """
        Args:
            days: Total duration in days
        """
        ...

    @classmethod
    def from_seconds(cls, seconds: float) -> TimeSpan:
        """
        Args:
            seconds: Total duration in seconds
        """
        ...

    @classmethod
    def from_minutes(cls, minutes: float) -> TimeSpan:
        """
        Args:
            minutes: Total duration in minutes
        """
        ...

    @classmethod
    def from_hours(cls, hours: float) -> TimeSpan:
        """
        Args:
            hours: Total duration in hours
        """
        ...

    def in_days(self) -> float:
        """
        Returns:
            Total duration in days
        """
        ...

    def in_seconds(self) -> float:
        """
        Returns:
            Total duration in seconds
        """
        ...

    def in_minutes(self) -> float:
        """
        Returns:
            Total duration in minutes
        """
        ...

    def in_hours(self) -> float:
        """
        Returns:
            Total duration in hours
        """
        ...

class TimeComponents:
    year: int
    """"""

    month: int
    """"""

    day: int
    """"""

    hour: int
    """"""

    minute: int
    """"""

    second: float
    """"""

    def __init__(
        self,
        year: int,
        month: int,
        day: int,
        hour: int,
        minute: int,
        second: float,
    ) -> None:
        """
        Initialize TimeComponents with the given year, month, day, hour, minute, and second.

        !!! warning
            It is the user's responsibility to input valid month/day combinations and time values.  This class does not
            perform validation.

        Args:
            year: Year (e.g., 2023)
            month: Month (1-12)
            day: Day (1-31)
            hour: Hour (0-23)
            minute: Minute (0-59)
            second: Second (0.0-59.999...)
        """

    def to_iso(self) -> str:
        """
        Returns:
            Epoch in ISO 8601 format (YYYY-MM-DDThh:mm:ss.sss)
        """
        ...

    @classmethod
    def from_iso(cls, iso_str: str) -> TimeComponents:
        """
        Args:
            iso_str: ISO 8601 formatted string (YYYY-MM-DDThh:mm:ss.sss)
        """
        ...

class Epoch:
    days_since_1950: float
    """
    Core floating epoch representation used by the SAAL binaries.
    """

    time_system: TimeSystem
    """"""

    day_of_year: float
    """Decimal day of the year (1-365.999...)"""

    def to_datetime(self) -> datetime:
        """
        Returns:
            Aware datetime object in UTC time system
        """
        ...

    @classmethod
    def from_datetime(cls, dt: datetime) -> Epoch:
        """
        Args:
            dt: Aware or naive datetime object (assumed to be UTC if naive)
        """
        ...

    @classmethod
    def now(cls) -> Epoch:
        """
        Returns:
            Current epoch in UTC time system
        """
        ...

    @classmethod
    def from_days_since_1950(cls, days: float, time_system: TimeSystem) -> Epoch:
        """
        Args:
            days: Days since 1 Jan 1956
            time_system: System used to calculate the number of days since 1 Jan 1956
        """
        ...

    @classmethod
    def from_iso(cls, iso_str: str, time_system: TimeSystem) -> Epoch:
        """
        Args:
            iso_str: ISO 8601 formatted string (YYYY-MM-DDThh:mm:ss.sss)
            time_system: System used to calculate the epoch
        """
        ...

    @classmethod
    def from_time_components(cls, components: TimeComponents, time_system: TimeSystem) -> Epoch:
        """
        Args:
            components: Epoch represented as individual components
            time_system: System used to calculate the epoch
        """
        ...

    @classmethod
    def from_dtg(cls, dtg: str, time_system: TimeSystem) -> Epoch:
        """
        Create an epoch from a standard SAAL DTG format.

        !!! note
            DTG formats include:

            - DTG 20: YYYY/DDD HHMM SS.SSS
            - DTG 19: YYYYMonDDHHMMSS.SSS
            - DTG 17: YYYY/DDD.DDDDDDDD
            - DTG 15: YYDDDHHMMSS.SSS

        Args:
            dtg: DTG formatted string (YYYY/DDD hhmm ss.sss)
            time_system: System used to calculate the epoch
        """
        ...

    def to_iso(self) -> str:
        """
        Returns:
            Epoch in ISO 8601 format (YYYY-MM-DDThh:mm:ss.sss)
        """
        ...

    def to_dtg_20(self) -> str:
        """
        Returns:
            Epoch in DTG 20 format (YYYY/DDD HHMM SS.SSS)
        """
        ...

    def to_dtg_19(self) -> str:
        """
        Returns:
            Epoch in DTG 19 format (YYYYMonDDHHMMSS.SSS)
        """
        ...

    def to_dtg_17(self) -> str:
        """
        Returns:
            Epoch in DTG 17 format (YYYY/DDD.DDDDDDDD)
        """
        ...

    def to_dtg_15(self) -> str:
        """
        Returns:
            Epoch in DTG 15 format (YYDDDHHMMSS.SSS)
        """
        ...

    def to_time_components(self) -> TimeComponents:
        """
        Returns:
            Epoch as individual components
        """
        ...

    def to_fk4_greenwich_angle(self) -> float:
        """
        Returns:
            Greenwich angle in radians using FK4 theory
        """
        ...

    def to_fk5_greenwich_angle(self) -> float:
        """
        Returns:
            Greenwich angle in radians using FK5 theory
        """
        ...

    def to_system(self, time_system: TimeSystem) -> Epoch:
        """
        Convert the epoch to a different time system.

        Args:
            time_system: Desired ouput system
        """
        ...

    def __add__(self, delta: TimeSpan) -> Epoch: ...
    @overload
    def __sub__(self, other: TimeSpan) -> Epoch: ...
    @overload
    def __sub__(self, other: Epoch) -> TimeSpan: ...
    def __gt__(self, other: Epoch) -> bool: ...
    def __lt__(self, other: Epoch) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __ge__(self, other: Epoch) -> bool: ...
    def __le__(self, other: Epoch) -> bool: ...



"""
Python test script for GPU-accelerated batch propagation

This script demonstrates the new batch propagation API exposed to Python,
including automatic GPU acceleration when beneficial.
"""

import sys

sys.path.insert(0, "target/debug")

try:
    from keplemon.propagation import BatchPropagator, PropagationBackend
    from keplemon.elements import TLE
    from keplemon.time import Epoch, TimeSpan
    from keplemon.bodies import Constellation
    from keplemon.enums import TimeSystem

except ImportError as e:
    print(f"Error importing keplemon: {e}")
    print("Make sure to build with: cargo build --features cuda,python")
    sys.exit(1)

# Test TLE lines
ISS_LINE1 = "1 25544U 98067A   24018.42784701  .00012593  00000+0  22161-3 0  9999"
ISS_LINE2 = "2 25544  51.6415 290.2344 0005730  47.3470  92.9368 15.50238117436068"

STARLINK_LINE1 = "1 44713U 19074A   24018.50000000  .00001749  00000+0  13559-3 0  9998"
STARLINK_LINE2 = "2 44713  53.0544 123.4567 0001234  45.6789  12.3456 15.06400000262500"


def test_tle_propagate_batch():
    """Test TLE.propagate_batch static method"""
    print("\n=== Testing TLE.propagate_batch() ===")

    # Create TLEs
    tle1 = TLE.from_lines(ISS_LINE1, ISS_LINE2)
    tle2 = TLE.from_lines(STARLINK_LINE1, STARLINK_LINE2)
    tles = [tle1, tle2]

    # Create epochs
    start = Epoch.from_iso("2024-01-18T12:00:00.000000Z", TimeSystem.UTC)
    epochs = [start + TimeSpan.from_hours(float(i)) for i in range(10)]

    print(f"  Propagating {len(tles)} TLEs to {len(epochs)} epochs...")

    # Propagate
    results = TLE.propagate_batch(tles, epochs)

    print(f"  ✓ Got {len(results)} satellites")
    print(f"  ✓ Each satellite has {len(results[0])} states")
    print(
        f"  ✓ First state position: [{results[0][0].position.x:.3f}, "
        f"{results[0][0].position.y:.3f}, {results[0][0].position.z:.3f}] km"
    )

    return results


def test_tle_propagate_to_epochs():
    """Test TLE.propagate_to_epochs instance method"""
    print("\n=== Testing TLE.propagate_to_epochs() ===")

    tle = TLE.from_lines(ISS_LINE1, ISS_LINE2)

    # Create many epochs to trigger GPU threshold
    start = Epoch.from_iso("2024-01-18T12:00:00.000000Z", TimeSystem.UTC)
    epochs = [start + TimeSpan.from_minutes(float(i)) for i in range(150)]

    print(f"  Propagating single TLE to {len(epochs)} epochs...")
    print("  (Should use GPU automatically with threshold=100)")

    states = tle.propagate_to_epochs(epochs)

    print(f"  ✓ Got {len(states)} states")

    # Check continuity
    for i in range(len(states) - 1):
        pos1 = states[i].position
        pos2 = states[i + 1].position
        dx = pos2.x - pos1.x
        dy = pos2.y - pos1.y
        dz = pos2.z - pos1.z
        dist = (dx**2 + dy**2 + dz**2) ** 0.5

        # With 1-minute spacing, ISS should move ~450 km
        if dist > 500:
            print(f"  ⚠ Warning: Large jump at step {i}: {dist:.1f} km")

    print("  ✓ States are continuous")

    return states


def test_batch_propagator():
    """Test BatchPropagator class directly"""
    print("\n=== Testing BatchPropagator ===")

    tle1 = TLE.from_lines(ISS_LINE1, ISS_LINE2)
    tle2 = TLE.from_lines(STARLINK_LINE1, STARLINK_LINE2)
    tles = [tle1, tle2]

    start = Epoch.from_iso("2024-01-18T12:00:00.000000Z", TimeSystem.UTC)
    epochs = [start + TimeSpan.from_hours(float(i) * 0.5) for i in range(20)]

    # Create propagator with explicit GPU backend (if available)
    propagator = BatchPropagator()

    gpu_available = propagator.is_gpu_available()
    print(f"  GPU available: {gpu_available}")

    if gpu_available:
        # Test with GPU
        propagator.set_backend(PropagationBackend.Gpu)
        print("  Propagating with GPU backend...")
        gpu_results = propagator.propagate_batch(tles, epochs)
        print(f"  ✓ GPU: {len(gpu_results)} satellites × {len(gpu_results[0])} epochs")

        # Test with CPU for comparison
        propagator.set_backend(PropagationBackend.Cpu)
        print("  Propagating with CPU backend...")
        cpu_results = propagator.propagate_batch(tles, epochs)
        print(f"  ✓ CPU: {len(cpu_results)} satellites × {len(cpu_results[0])} epochs")

        # Compare results
        max_diff = 0.0
        for sat_idx in range(len(tles)):
            for time_idx in range(len(epochs)):
                cpu_pos = cpu_results[sat_idx][time_idx].position
                gpu_pos = gpu_results[sat_idx][time_idx].position

                dx = gpu_pos.x - cpu_pos.x
                dy = gpu_pos.y - cpu_pos.y
                dz = gpu_pos.z - cpu_pos.z
                diff = (dx**2 + dy**2 + dz**2) ** 0.5

                if diff > max_diff:
                    max_diff = diff

        max_diff_m = max_diff * 1e3
        print(f"  ✓ Max CPU vs GPU difference: {max_diff_m:.1f} m")

        if max_diff < 0.1:  # 100m tolerance
            print("  ✓ Results match within tolerance")
        else:
            print("  ⚠ Warning: Difference exceeds 100m")
    else:
        # CPU only
        propagator.set_backend(PropagationBackend.Cpu)
        print("  Propagating with CPU backend (GPU not available)...")
        results = propagator.propagate_batch(tles, epochs)
        print(f"  ✓ CPU: {len(results)} satellites × {len(results[0])} epochs")


def test_constellation_batch():
    """Test Constellation batch methods"""
    print("\n=== Testing Constellation batch methods ===")

    tle1 = TLE.from_lines(ISS_LINE1, ISS_LINE2)
    tle2 = TLE.from_lines(STARLINK_LINE1, STARLINK_LINE2)

    constellation = Constellation()
    constellation.add("ISS", tle1.to_satellite())
    constellation.add("Starlink", tle2.to_satellite())

    print(f"  Constellation has {constellation.get_count()} satellites")

    # Test batch ephemeris
    start = Epoch.from_iso("2024-01-18T12:00:00.000000Z", TimeSystem.UTC)
    end = start + TimeSpan.from_hours(2.0)
    step = TimeSpan.from_minutes(10.0)

    print("  Computing batch ephemeris...")

    # Try with Auto backend
    results = constellation.get_batch_ephemeris(start, end, step, PropagationBackend.Auto)

    print(f"  ✓ Got results for {len(results)} satellites")

    for sat_id, states in results.items():
        valid_states = [s for s in states if s is not None]
        print(f"  ✓ {sat_id}: {len(valid_states)}/{len(states)} valid states")

    # Test GPU availability
    gpu_available = Constellation.is_gpu_available()
    print(f"  ✓ GPU available (static method): {gpu_available}")


def main():
    print("=" * 60)
    print("GPU-Accelerated Batch Propagation Python Tests")
    print("=" * 60)

    try:
        test_tle_propagate_batch()
        test_tle_propagate_to_epochs()
        test_batch_propagator()
        # test_constellation_batch()  # Commented out - needs to_satellite() method

        print("\n" + "=" * 60)
        print("✓ All tests passed!")
        print("=" * 60)

    except Exception as e:
        print(f"\n✗ Test failed with error: {e}")
        import traceback

        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations
import click
from pathlib import Path
from importlib import resources
from keplemon.time import request_time_constants_update


@click.command()
@click.option(
    "--update-eop",
    help="Update time constants and EOP data (global or path/to/output/file)",
    type=click.Path(exists=False),
)
def cli(update_eop: Path | None) -> None:
    if update_eop is not None:
        if update_eop == "global":
            time_constants_path = resources.files("keplemon") / "time_constants.dat"
        print("Requesting time constants and EOP data from USNO...")
        request_time_constants_update(str(time_constants_path))
        print(f"Updated time constants at {update_eop}")



# flake8: noqa
from __future__ import annotations

import os
import logging
from importlib import resources

_logger = logging.getLogger(__name__)


def _set_asset_directory() -> None:

    asset_dir = os.getenv("SAAL_ASSET_DIRECTORY")
    if asset_dir is None:
        pkg_dir = resources.files("keplemon")
        _logger.debug("Setting SAAL_ASSET_DIRECTORY to %s", pkg_dir)
        os.environ.setdefault("SAAL_ASSET_DIRECTORY", str(pkg_dir))
    elif not os.path.exists(asset_dir):
        raise FileNotFoundError(f"SAAL_ASSET_DIRECTORY '{asset_dir}' does not exist.")


_set_asset_directory()

from keplemon._keplemon import (  # type: ignore
    get_thread_count,
    set_thread_count,
)

__all__ = [
    "get_thread_count",
    "set_thread_count",
]



# flake8: noqa
from pathlib import Path

def get_thread_count() -> int:
    """
    Returns:
        Number of cores allocated for use by KepLemon
    """
    ...

def set_thread_count(n: int) -> None:
    """
    Set the number of cores allocated for use by KepLemon

    !!! warning
        This function must be called before any other functions in the library

    Args:
        n: Number of cores to allocate
    """
    ...

#: Path to the time constants file
TIME_CONSTANTS_PATH: Path
"""
Path to the default time constants file required by the SAAL binaries

!!! warning
    This path should never be modified and is only exposed to allow inspection of current data.
"""

#: Path to the parent directory of the package
PACKAGE_DIRECTORY: Path
"""Path to the parent directory of the package"""

#: Path to the assets directory containing supporting data files for the package
ASSETS_DIRECTORY: Path
"""Path to the assets directory containing supporting data files for the package"""

def set_license_file_path(path: str) -> None:
    """
    Set the path to the license file required by the SAAL binaries

    Args:
        path: Path to the SGP4 license file
    """
    ...

def get_license_file_path() -> str:
    """
    Returns:
        Path to the SGP4 license file
    """
    ...



from keplemon._keplemon.propagation import (  # type: ignore
    ForceProperties,
    InertialPropagator,
    SGP4Output,
    b_star_to_drag_coefficient,
    drag_coefficient_to_b_star,
    BatchPropagator,
    PropagationBackend,
)

__all__ = [
    "ForceProperties",
    "InertialPropagator",
    "SGP4Output",
    "b_star_to_drag_coefficient",
    "drag_coefficient_to_b_star",
    "BatchPropagator",
    "PropagationBackend",
]



# flake8: noqa
from typing import Iterator, Optional

from keplemon.elements import (
    TLE,
    CartesianState,
    Ephemeris,
    KeplerianState,
    GeodeticPosition,
    OrbitPlotData,
    TopocentricElements,
    RelativeState,
    BoreToBodyAngles,
)
from keplemon.catalogs import TLECatalog
from keplemon.time import Epoch, TimeSpan
from keplemon.events import CloseApproach, CloseApproachReport, HorizonAccessReport, FieldOfViewReport, ManeuverEvent, ManeuverReport, ProximityReport, UCTValidityReport
from keplemon.propagation import ForceProperties
from keplemon.enums import ReferenceFrame
from keplemon.estimation import CollectionAssociationReport, Observation, ObservationAssociation, ObservationCollection, ObservationResidual

class Satellite:

    id: str
    """Unique identifier for the satellite."""

    norad_id: int
    """Number corresponding to the satellite's NORAD catalog ID.
    """

    force_properties: ForceProperties
    """Force properties of the satellite used for propagation"""

    name: Optional[str]
    """Human-readable name of the satellite"""

    keplerian_state: Optional[KeplerianState]
    """Keplerian state of the satellite at the epoch of the TLE, if available"""

    geodetic_position: Optional[GeodeticPosition]
    """Geodetic position of the satellite at the epoch of the TLE, if available"""

    def __init__(self) -> None: ...
    @classmethod
    def from_tle(cls, tle: TLE) -> Satellite:
        """
        Instantiate a Satellite from a legacy TLE

        Args:
            tle: Two-line element set for the satellite
        """
        ...

    def get_close_approach(
        self,
        other: Satellite,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
    ) -> Optional[CloseApproach]: ...
    def get_proximity_report(
        self,
        other: Satellite,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
    ) -> Optional[ProximityReport]:
        """
        Check if this satellite stays within a distance threshold of another satellite.

        Args:
            other: Satellite to compare against
            start: UTC epoch of the start of the report
            end: UTC epoch of the end of the report
            distance_threshold: Maximum distance threshold in **_kilometers_**

        Returns:
            Proximity report with a single event if satellites stay within threshold,
            or an empty report if the threshold is exceeded at any point
        """
        ...
    def get_maneuver_event(
        self,
        future_sat: Satellite,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
        velocity_threshold: float,
    ) -> Optional[ManeuverEvent]:
        """
        Detect a maneuver by comparing this satellite's state with a future state.

        Args:
            future_sat: Satellite with a future epoch state to compare against
            start: UTC epoch of the start of the search window
            end: UTC epoch of the end of the search window
            distance_threshold: Distance threshold for matching in **_kilometers_**
            velocity_threshold: Velocity threshold for maneuver detection in **_meters per second_**

        Returns:
            ManeuverEvent if a maneuver is detected, None otherwise
        """
        ...
    def get_ephemeris(
        self,
        start: Epoch,
        end: Epoch,
        step: TimeSpan,
    ) -> Ephemeris: ...
    def get_state_at_epoch(self, epoch: Epoch) -> CartesianState: ...
    def to_tle(self) -> Optional[TLE]:
        """
        Returns:
            Satellite as a two-line element set or None if no state is loaded

        """
        ...

    def get_relative_state_at_epoch(self, other: Satellite, epoch: Epoch) -> Optional[RelativeState]:
        """
        Calculate the relative state between this satellite and another satellite at a given epoch.

        Args:
            other: Secondary satellite to calculate the relative state against
            epoch: UTC epoch at which the relative state will be calculated
        """
        ...

    def get_body_angles_at_epoch(self, other: Satellite, epoch: Epoch) -> Optional[BoreToBodyAngles]:
        """
        Calculate the bore-to-body angles between this satellite and another satellite at a given epoch.

        Args:
            other: Secondary satellite to calculate the bore-to-body angles against
            epoch: UTC epoch at which the bore-to-body angles will be calculated
        """
        ...

    def get_plot_data(self, start: Epoch, end: Epoch, step: TimeSpan) -> Optional[OrbitPlotData]: ...
    def get_observatory_access_report(
        self,
        observatories: list[Observatory],
        start: Epoch,
        end: Epoch,
        min_el: float,
        min_duration: TimeSpan,
    ) -> Optional[HorizonAccessReport]:
        """
        Calculate horizon access from multiple observatories to this satellite.

        Args:
            observatories: List of observatories to check for horizon access
            start: UTC epoch of the start of the report
            end: UTC epoch of the end of the report
            min_el: Minimum elevation angle in **_degrees_**
            min_duration: Minimum duration of access

        Returns:
            Horizon access report containing accesses from all observatories to the satellite,
            or None if the satellite ephemeris cannot be generated
        """
        ...

    def get_associations(
        self, collections: list[ObservationCollection]
    ) -> list[ObservationAssociation]:
        """
        Find observation associations for this satellite across multiple observation collections.

        Args:
            collections: List of observation collections to search for associations

        Returns:
            List of observation associations where this satellite matches observations
        """
        ...

    def get_rms(self, obs: list[Observation]) -> float:
        """
        Calculate the root mean squared position error between observations and the satellite state.

        Uses interpolated states from cached ephemeris when available for better performance.

        Args:
            obs: List of observations to compare against

        Returns:
            Root mean squared position error in **_kilometers_**

        Raises:
            ValueError: If no valid residuals could be computed
        """
        ...

    def get_residuals(self, obs: list[Observation]) -> list[ObservationResidual]:
        """
        Calculate position residuals between observations and the satellite state.

        Uses interpolated states from cached ephemeris when available for better performance.

        Args:
            obs: List of observations to compare against

        Returns:
            List of observation residuals

        Raises:
            ValueError: If no valid residuals could be computed
        """
        ...

class Constellation:

    count: int
    """Number of satellites in the constellation"""

    name: Optional[str]
    """Human-readable name of the constellation"""

    def __init__(self) -> None: ...
    def get_plot_data(self, start: Epoch, end: Epoch, step: TimeSpan) -> dict[str, OrbitPlotData]: ...
    @classmethod
    def from_tle_catalog(cls, tle_catalog: TLECatalog) -> Constellation:
        """
        Instantiate a Constellation from a TLE catalog

        Args:
            tle_catalog: TLE catalog for the constellation
        """
        ...

    def get_states_at_epoch(self, epoch: Epoch) -> dict[int, CartesianState]:
        """
        Args:
            epoch: UTC epoch at which the states will be calculated

        Returns:
            (satellite_id, state) dictionary for the constellation at the given epoch
        """
        ...

    def get_ephemeris(
        self,
        start: Epoch,
        end: Epoch,
        step: TimeSpan,
    ) -> dict[str, Ephemeris]:
        """
        Args:
            start: UTC epoch of the start of the ephemeris
            end: UTC epoch of the end of the ephemeris
            step: Time step for the ephemeris

        Returns:
            (satellite_id, ephemeris) dictionary for the constellation
        """
        ...

    def get_ca_report_vs_one(
        self,
        other: Satellite,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
    ) -> CloseApproachReport:
        """
        Calculate close approaches between the constellation and a given satellite.

        Args:
            other: Satellite to compare against
            start: UTC epoch of the start of the close approach report
            end: UTC epoch of the end of the close approach report
            distance_threshold: Distance threshold for close approach screening in **_kilometers_**

        Returns:
            Close approach report for the constellation vs. the given satellite
        """
        ...

    def get_ca_report_vs_many(
        self,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
    ) -> CloseApproachReport:
        """
        Calculate close approaches among satellites in the calling constellation.

        !!! warning
            This is a long-running operation when the constellation is large.

        Args:
            start: UTC epoch of the start of the close approach report
            end: UTC epoch of the end of the close approach report
            distance_threshold: Distance threshold for close approach screening in **_kilometers_**

        Returns:
            Close approach report for the constellation vs. all other satellites
        """
        ...

    def get_proximity_report_vs_one(
        self,
        other: Satellite,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
    ) -> ProximityReport:
        """
        Check which constellation satellites stay within a distance threshold of a given satellite.

        Args:
            other: Satellite to compare against
            start: UTC epoch of the start of the report
            end: UTC epoch of the end of the report
            distance_threshold: Maximum distance threshold in **_kilometers_**

        Returns:
            Proximity report containing events for satellite pairs that stay within threshold
        """
        ...

    def get_proximity_report_vs_many(
        self,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
    ) -> ProximityReport:
        """
        Check which satellite pairs in the constellation stay within a distance threshold.

        Args:
            start: UTC epoch of the start of the report
            end: UTC epoch of the end of the report
            distance_threshold: Maximum distance threshold in **_kilometers_**

        Returns:
            Proximity report containing events for satellite pairs that stay within threshold
        """
        ...

    def get_uct_validity(
        self,
        uct: Satellite,
        all_collections: list[ObservationCollection],
        orphan_collections: list[ObservationCollection],
    ) -> UCTValidityReport:
        """
        Analyze the validity of a UCT (Uncorrelated Track) against the constellation.

        This method uses cached ephemeris to analyze:
        - Observation associations with orphan observations
        - Proximity events with approved satellites (possible cross-tags)
        - Close approaches with approved satellites (possible maneuver origins)

        The analysis window is automatically determined based on the UCT's orbital period
        and the constellation's cached ephemeris bounds.

        Args:
            uct: The UCT satellite to analyze
            all_collections: All observation collections for visibility checks
            orphan_collections: Observation collections of orphan observations

        Returns:
            UCT validity report with analysis results

        Raises:
            ValueError: If the UCT satellite has no valid orbit state
        """
        ...

    def get_maneuver_events(
        self,
        future_sats: Constellation,
        start: Epoch,
        end: Epoch,
        distance_threshold: float,
        velocity_threshold: float,
    ) -> ManeuverReport:
        """
        Detect maneuvers by comparing current satellite states with future states.

        Matches satellites by ID between the two constellations and detects maneuvers
        where the velocity difference exceeds the threshold.

        Args:
            future_sats: Constellation with future epoch states to compare against
            start: UTC epoch of the start of the search window
            end: UTC epoch of the end of the search window
            distance_threshold: Distance threshold for matching in **_kilometers_**
            velocity_threshold: Velocity threshold for maneuver detection in **_meters per second_**

        Returns:
            ManeuverReport containing all detected maneuvers
        """
        ...

    def __getitem__(self, satellite_id: str) -> Satellite: ...
    def __setitem__(self, satellite_id: str, sat: Satellite) -> None: ...
    def __delitem__(self, satellite_id: str) -> None: ...
    def __iter__(self) -> Iterator[str]: ...
    def __contains__(self, key: str) -> bool: ...
    def __len__(self) -> int: ...
    def keys(self) -> list[str]: ...
    def get_horizon_access_report(
        self,
        site: Observatory,
        start: Epoch,
        end: Epoch,
        min_el: float,
        min_duration: TimeSpan,
    ) -> HorizonAccessReport:
        """
        Calculate horizon access to a given observatory.

        Args:
            site: Observatory to check for horizon access
            start: UTC epoch of the start of the report
            end: UTC epoch of the end of the report
            min_el: Minimum elevation angle in **_degrees_**
            min_duration: Minimum duration of access

        Returns:
            Horizon access report for the constellation from the observatory
        """
        ...

    def cache_ephemeris(self, start: Epoch, end: Epoch, step: TimeSpan, purge_on_fail: bool = False) -> None:
        """
        Cache ephemeris for all satellites in the constellation.

        This pre-computes and caches ephemeris data for each satellite, enabling
        faster interpolation-based state lookups via interpolate_state_at_epoch.

        Args:
            start: Start epoch for ephemeris caching
            end: End epoch for ephemeris caching
            step: Time step between cached states
            purge_on_fail: Remove satellites that fail to build ephemeris
        """
        ...

    def get_association_reports(
        self, collections: list[ObservationCollection]
    ) -> list[CollectionAssociationReport]:
        """
        Get association reports for multiple observation collections.

        Args:
            collections: List of observation collections to find associations for

        Returns:
            List of CollectionAssociationReport, one per input collection
        """
        ...

class Sensor:
    """
    Args:
        name: Identifier of the sensor
        angular_noise: Angular noise in **_degrees_**
    """

    id: str
    """Unique identifier for the sensor."""
    name: Optional[str]
    angular_noise: float
    range_noise: Optional[float]
    """Range noise in **_kilometers_**"""

    range_rate_noise: Optional[float]
    """Range rate noise in **_kilometers per second_**"""

    angular_rate_noise: Optional[float]
    """Angular rate noise in **_degrees per second_**"""
    def __init__(self, angular_noise: float) -> None: ...

class Observatory:
    """
    Args:
        latitude: Latitude in **_degrees_**
        longitude: Longitude in **_degrees_**
        altitude: Altitude in **_kilometers_**
    """

    name: str
    id: str
    """Unique identifier for the observatory."""
    latitude: float
    longitude: float
    altitude: float
    sensors: list[Sensor]
    """List of sensors at the observatory"""
    def __init__(
        self,
        latitude: float,
        longitude: float,
        altitude: float,
    ) -> None: ...
    def get_state_at_epoch(self, epoch: Epoch) -> CartesianState:
        """
        Args:
            epoch: UTC epoch of the state

        Returns:
            TEME Cartesian state of the observatory in **_kilometers_** and **_kilometers per second_**
        """
        ...

    @classmethod
    def from_cartesian_state(cls, state: CartesianState) -> Observatory:
        """
        Create an observatory from a Cartesian state.

        Args:
            state: Cartesian state of the observatory

        """
        ...

    def get_theta(self, epoch: Epoch) -> float:
        """
        Calculate the Greenwich angle plus the observatory longitude at a given epoch.

        Args:
            epoch: UTC epoch for the calculation

        Returns:
            Greenwich angle plus the observatory longitude in **_radians_**
        """
        ...

    def get_horizon_access_report(
        self,
        satellite: Satellite,
        start: Epoch,
        end: Epoch,
        min_el: float,
        min_duration: TimeSpan,
    ) -> HorizonAccessReport:
        """
        Calculate horizon access for a satellite from the observatory.

        Args:
            satellite: Satellite to check for horizon access
            start: UTC epoch of the start of the report
            end: UTC epoch of the end of the report
            min_el: Minimum elevation angle in **_degrees_**
            min_duration: Minimum duration of access in **_seconds_**

        Returns:
            Horizon access report for the satellite from the observatory
        """
        ...

    def get_field_of_view_report(
        self,
        epoch: Epoch,
        sensor_direction: TopocentricElements,
        angular_threshold: float,
        sats: Constellation,
        reference_frame: ReferenceFrame,
    ) -> FieldOfViewReport:
        """
        Calculate satellites in the field of view from a given time and direction.

        Args:
            epoch: UTC epoch of the report
            sensor_direction: Topocentric direction the sensor is pointing
            angular_threshold: Angular threshold in **_degrees_**
            sats: Constellation of satellites to check for being in the field of view
            reference_frame: Reference frame of the output direction elements
        """
        ...

    def get_topocentric_to_satellite(
        self,
        epoch: Epoch,
        sat: Satellite,
        reference_frame: ReferenceFrame,
    ) -> TopocentricElements:
        """
        Get the topocentric elements of a satellite as seen from the observatory.
        Args:
            epoch: UTC epoch of the observation
            sat: Satellite to observe
            reference_frame: Reference frame of the output direction elements
        """
        ...



from keplemon._keplemon.enums import (  # type: ignore
    TimeSystem,
    Classification,
    KeplerianType,
    ReferenceFrame,
    CovarianceType,
    AssociationConfidence,
    GeodeticModel,
    MeanEquinox,
    UCTObservability,
    UCTValidity,
)

__all__ = [
    "TimeSystem",
    "Classification",
    "KeplerianType",
    "ReferenceFrame",
    "CovarianceType",
    "AssociationConfidence",
    "GeodeticModel",
    "MeanEquinox",
    "UCTObservability",
    "UCTValidity",
]



from keplemon._keplemon.events import (  # type: ignore
    CloseApproach,
    CloseApproachReport,
    HorizonAccess,
    HorizonAccessReport,
    FieldOfViewCandidate,
    FieldOfViewReport,
    ManeuverEvent,
    ManeuverReport,
    ProximityEvent,
    ProximityReport,
    UCTValidityReport,
)

__all__ = [
    "CloseApproach",
    "CloseApproachReport",
    "HorizonAccess",
    "HorizonAccessReport",
    "FieldOfViewCandidate",
    "FieldOfViewReport",
    "ManeuverEvent",
    "ManeuverReport",
    "ProximityEvent",
    "ProximityReport",
    "UCTValidityReport",
]



from keplemon._keplemon.estimation import (  # type: ignore
    Observation,
    ObservationResidual,
    BatchLeastSquares,
    Covariance,
    ObservationAssociation,
    ObservationCollection,
    CollectionAssociationReport,
)

__all__ = [
    "Observation",
    "ObservationResidual",
    "BatchLeastSquares",
    "Covariance",
    "ObservationAssociation",
    "ObservationCollection",
    "CollectionAssociationReport",
]



# flake8: noqa
from typing import Optional

from keplemon.elements import TLE, OrbitPlotData

class TLECatalog:
    count: int
    name: Optional[str]
    def __init__(self) -> None: ...
    @classmethod
    def from_tle_file(cls, filename: str) -> TLECatalog: ...
    def add(self, tle: TLE) -> None: ...
    def get(self, satellite_id: str) -> TLE: ...
    def remove(self, satellite_id: str) -> None: ...
    def keys(self) -> list[str]: ...
    def get_count(self) -> int: ...
    def clear(self) -> None: ...
    def __getitem__(self, satellite_id: str) -> TLE: ...
    def get_plot_data(self) -> OrbitPlotData: ...
    def fit_best_tle(
        self,
        srp_coefficient: Optional[float] = None,
        drag_coefficient: Optional[float] = None,
    ) -> TLE:
        """
        Fit the best TLE from all TLEs in this catalog using batch least squares.

        Uses the first TLE as the a priori estimate and fits to all TLE observations.
        Automatically enables drag and SRP estimation based on orbital altitude,
        unless explicit coefficients are provided.

        Args:
            srp_coefficient: Optional SRP coefficient. If provided, SRP will not be estimated.
            drag_coefficient: Optional drag coefficient. If provided, drag will not be estimated.

        Returns:
            The best-fit TLE
        """
        ...



# flake8: noqa
from __future__ import annotations
from typing import Optional

from keplemon.time import Epoch
from keplemon.enums import Classification, KeplerianType, ReferenceFrame
from keplemon.estimation import Observation
from keplemon.events import CloseApproach
from keplemon.bodies import Observatory

class RelativeState:
    epoch: Epoch
    position: CartesianVector
    velocity: CartesianVector
    origin_satellite_id: str
    secondary_satellite_id: str

class BoreToBodyAngles:
    earth_angle: float
    sun_angle: float
    moon_angle: float

class OrbitPlotState:
    epoch: Epoch
    latitude: float
    longitude: float
    altitude: float
    semi_major_axis: float
    eccentricity: float
    inclination: float
    raan: float
    radius: float
    apogee_radius: float
    perigee_radius: float

class OrbitPlotData:
    satellite_id: str
    epochs: list[str]
    semi_major_axes: list[float]
    eccentricities: list[float]
    inclinations: list[float]
    raans: list[float]
    radii: list[float]
    apogee_radii: list[float]
    perigee_radii: list[float]
    latitudes: list[float]
    longitudes: list[float]
    altitudes: list[float]

class GeodeticPosition:
    """
    Args:
        latitude: Latitude in **_degrees_**
        longitude: Longitude in **_degrees_**
        altitude: Altitude in **_kilometers_**
    """

    latitude: float
    """Latitude in **_degrees_**"""

    longitude: float
    """Longitude in **_degrees_**"""

    altitude: float
    """Altitude in **_kilometers_**"""

    def __init__(self, latitude: float, longitude: float, altitude: float) -> None: ...

class HorizonElements:
    """
    Args:
        range: Range in **_kilometers_**
        az: Azimuth in **_degrees_**
        el: Elevation in **_degrees_**
        range_rate: Range rate in **_kilometers per second_**
        az_rate: Azimuth rate in **_degrees per second_**
        el_rate: Elevation rate in **_degrees per second_**
    """

    range: Optional[float]
    azimuth: float
    elevation: float
    range_rate: Optional[float]
    azimuth_rate: Optional[float]
    elevation_rate: Optional[float]

    def __init__(
        self,
        azimuth: float,
        elevation: float,
    ) -> None: ...

class HorizonState:
    """
    Args:
        epoch: UTC epoch of the state
        elements: HorizonElements of the state
    """

    epoch: Epoch
    """UTC epoch of the state"""

    elements: HorizonElements
    """Horizon elements of the state"""

    range: Optional[float]
    """Range in **_kilometers_**"""

    azimuth: float
    """Azimuth in **_degrees_**"""

    elevation: float
    """Elevation in **_degrees_**"""

    range_rate: Optional[float]
    """Range rate in **_kilometers per second_**"""

    azimuth_rate: Optional[float]
    """Azimuth rate in **_degrees per second_**"""

    elevation_rate: Optional[float]
    """Elevation rate in **_degrees per second_**"""

    def __init__(self, epoch: Epoch, elements: HorizonElements) -> None: ...
    @classmethod
    def from_topocentric_state(cls, state: TopocentricState, observer: Observatory) -> HorizonState: ...

class KeplerianElements:
    """
    Args:
        semi_major_axis: Average distance from the central body in **_kilometers_**
        eccentricity: Eccentricity of the orbit
        inclination: Inclination of the orbit in **_degrees_**
        raan: Right Ascension of Ascending Node in **_degrees_**
        argument_of_perigee: Argument of Perigee in **_degrees_**
        mean_anomaly: Mean Anomaly in **_degrees_**
    """

    semi_major_axis: float
    eccentricity: float
    inclination: float
    raan: float
    argument_of_perigee: float
    mean_anomaly: float

    def __init__(
        self,
        semi_major_axis: float,
        eccentricity: float,
        inclination: float,
        raan: float,
        argument_of_perigee: float,
        mean_anomaly: float,
    ) -> None: ...

class EquinoctialElements:
    """
    Args:
        a_f: Equinoctial element a_f
        a_g: Equinoctial element a_g
        chi: Equinoctial element chi
        psi: Equinoctial element psi
        mean_longitude: Mean longitude in **_degrees_**
        mean_motion: Mean motion in **_revolutions per day_**
    """

    a_f: float
    a_g: float
    chi: float
    psi: float
    mean_longitude: float
    mean_motion: float

    def __init__(
        self,
        a_f: float,
        a_g: float,
        chi: float,
        psi: float,
        mean_longitude: float,
        mean_motion: float,
    ) -> None: ...

    def to_keplerian(self) -> KeplerianElements: ...

class TLE:

    norad_id: int
    """NORAD catalog ID of the satellite"""

    satellite_id: str
    """"""

    name: str
    """"""

    inclination: float
    """Inclination of the orbit in **_degrees_**"""

    eccentricity: float
    """"""

    raan: float
    """Right Ascension of Ascending Node in **_degrees_**"""

    argument_of_perigee: float
    """Argument of Perigee in **_degrees_**"""

    mean_anomaly: float
    """Mean Anomaly in **_degrees_**"""

    mean_motion: float
    """Mean motion in **_revolutions per day_**"""

    type: KeplerianType
    """"""

    b_star: float
    """"""

    mean_motion_dot: float
    """"""

    mean_motion_dot_dot: float
    """"""

    agom: float
    """"""

    b_term: float
    """"""

    epoch: Epoch
    """UTC epoch of the state"""

    classification: Classification
    """"""

    designator: str
    """8-character identifier of the satellite"""

    cartesian_state: CartesianState
    """TEME cartesian state of the TLE at epoch"""

    semi_major_axis: float
    """Average distance from the central body in **_kilometers_**
    
    !!! note
        This is always calculated using Brouwer mean motion and will differ slightly from Kozai-computed SMA.
    """

    apoapsis: float
    """Apoapsis radius in **_kilometers_**"""

    periapsis: float
    """Periapsis radius in **_kilometers_**"""

    @classmethod
    def from_lines(cls, line_1: str, line_2: str, line_3: Optional[str] = None) -> TLE:
        """
        Create a TLE object using strings in 2 or 3 line format
        """
        ...

    @property
    def lines(self) -> tuple[str, str]:
        """
        !!! note
            If the TLE was created in the 3LE format, only lines 2 and 3 will be returned.  The name must be accessed
            using the `name` property.

        Returns:
            Tuple of strings in 2 line format
        """
        ...

    def get_state_at_epoch(self, epoch: Epoch) -> CartesianState:
        """
        Get a Cartesian state at a specific epoch by propagating the TLE.

        Args:
            epoch: The epoch at which to get the state

        Returns:
            CartesianState in TEME frame at the specified epoch
        """
        ...

    def get_keplerian_state_at_epoch(self, epoch: Epoch) -> KeplerianState:
        """
        Get a Keplerian state at a specific epoch by propagating the TLE.

        Returns mean elements appropriate to the TLE type (Kozai for SGP4, Brouwer for XP).

        Args:
            epoch: The epoch at which to get the state

        Returns:
            KeplerianState in TEME frame at the specified epoch
        """
        ...

    def get_observation_at_epoch(self, epoch: Epoch) -> Observation:
        """
        Get an observation at a specific epoch.

        Creates an observation from the TLE's propagated state at the given epoch,
        useful for batch least squares fitting of multiple TLEs.

        Args:
            epoch: The epoch at which to generate the observation

        Returns:
            Observation derived from this TLE at the given epoch
        """
        ...

class SphericalVector:
    """
    !!! note
        The range units can be disregarded if this class is not being used for astrodynamic transforms.

    Args:
        range: distance from the origin in **_kilometers_**
        right_ascension: right ascension in **_degrees_**
        declination: declination in **_degrees_**
    """

    range: float
    right_ascension: float
    declination: float
    def __init__(self, range: float, right_ascension: float, declination: float) -> None: ...
    def to_cartesian(self) -> CartesianVector:
        """
        Returns:
            Cartesian vector in **_kilometers_**.
        """
        ...

class CartesianVector:
    """
    Args:
        x: x coordinate
        y: y coordinate
        z: z coordinate
    """

    x: float
    y: float
    z: float
    magnitude: float
    """"""

    def __init__(self, x: float, y: float, z: float) -> None: ...
    def distance(self, other: CartesianVector) -> float:
        """
        !!! note
            Take care to use consistent units with this function.

        Returns:
            Distance between two Cartesian vectors
        """
        ...

    def to_spherical(self) -> SphericalVector:
        """
        Returns:
            Spherical representation of the point
        """
        ...

    def __add__(self, other: CartesianVector) -> CartesianVector: ...
    def __sub__(self, other: CartesianVector) -> CartesianVector: ...
    def angle(self, other: CartesianVector) -> float:
        """
        Returns:
            Angle between two Cartesian vectors in **_radians_**.
        """
        ...

class CartesianState:
    """State represented as x, y, z and vx, vy, vz in a given reference frame.

    Args:
        epoch: UTC epoch of the state
        position: Cartesian position vector in kilometers
        velocity: Cartesian velocity vector in kilometers per second
        frame: reference frame of the state
    """

    position: CartesianVector
    """Position of the state in kilometers"""

    velocity: CartesianVector
    """Velocity of the state in kilometers per second"""

    epoch: Epoch
    """UTC epoch of the state"""

    frame: ReferenceFrame
    """Current reference frame of the state"""

    def __init__(
        self, epoch: Epoch, position: CartesianVector, velocity: CartesianVector, frame: ReferenceFrame
    ) -> None: ...
    def to_keplerian(self) -> KeplerianState:
        """Convert the Cartesian state to osculating Keplerian elements"""
        ...

    def to_frame(self, frame: ReferenceFrame) -> CartesianState:
        """

        Args:
            frame: reference frame of the output state

        Returns:
            CartesianState: Cartesian state in the new frame"""
        ...

class KeplerianState:
    """Orbit represented as Keplerian elements in a given reference frame.

    Args:
        epoch: UTC epoch of the state
        elements: Keplerian elements of the state
        frame: reference frame of the state
        keplerian_type: type of the Keplerian elements
    """

    epoch: Epoch
    frame: ReferenceFrame
    type: KeplerianType
    semi_major_axis: float
    """Average distance from the central body in **_kilometers_**"""

    eccentricity: float
    """Eccentricity of the orbit"""

    inclination: float
    """Inclination of the orbit in **_degrees_**"""

    raan: float
    """Right Ascension of Ascending Node in **_degrees_**"""

    argument_of_perigee: float
    """Argument of Perigee in **_degrees_**"""

    mean_anomaly: float
    """Mean Anomaly in **_degrees_**"""

    mean_motion: float
    """Mean motion in **_revolutions per day_**"""

    apoapsis: float
    """Furthest point from the central body in **_kilometers_**"""

    periapsis: float
    """Closest point to the central body in **_kilometers_**"""

    def __init__(
        self,
        epoch: Epoch,
        elements: KeplerianElements,
        frame: ReferenceFrame,
        keplerian_type: KeplerianType,
    ) -> None: ...
    def to_cartesian(self) -> CartesianState:
        """
        Returns:
            Cartesian state in **_kilometers_** and **_kilometers per second_**.
        """
        ...

    def to_frame(self, frame: ReferenceFrame) -> KeplerianState:
        """
        Args:
            frame: reference frame of the output state

        Returns:
            Keplerian state in the new frame"""
        ...

class Ephemeris:
    def get_close_approach(
        self,
        other: Ephemeris,
        distance_threshold: float,
    ) -> CloseApproach: ...

class TopocentricElements:
    """
    Args:
        right_ascension: TEME right ascension in **_degrees_**
        declination: TEME declination in **_degrees_**
    """

    range: Optional[float]
    """Range in **_kilometers_**"""

    right_ascension: float
    declination: float
    range_rate: Optional[float]
    """Range rate in **_kilometers per second_**"""

    right_ascension_rate: Optional[float]
    """Right ascension rate in **_degrees per second**"""

    declination_rate: Optional[float]
    """Declination rate in **_degrees per second**"""

    def __init__(self, right_ascension: float, declination: float) -> None: ...
    @classmethod
    def from_j2000(cls, epoch: Epoch, ra: float, dec: float) -> TopocentricElements:
        """
        Args:
            epoch: UTC epoch of the angles
            ra: J2000 right ascension in **_degrees_**
            dec: J2000 declination in **_degrees_**
        """
        ...

class TopocentricState:
    """
    Args:
        epoch: UTC epoch of the state
        elements: TopocentricElements of the state
    """

    epoch: Epoch
    """UTC epoch of the state"""

    elements: TopocentricElements
    """Topocentric elements of the state"""

    range: Optional[float]
    """Range in **_kilometers_**"""

    right_ascension: float
    """TEME right ascension in **_degrees_**"""

    declination: float
    """TEME declination in **_degrees_**"""

    range_rate: Optional[float]
    """Range rate in **_kilometers per second_**"""

    right_ascension_rate: Optional[float]
    """Right ascension rate in **_degrees per second**"""

    declination_rate: Optional[float]
    """Declination rate in **_degrees per second**"""

    def __init__(self, epoch: Epoch, elements: TopocentricElements) -> None: ...
    @classmethod
    def from_horizon_state(cls, horizon_state: HorizonState, observer: Observatory) -> TopocentricState:
        """
        Args:
            horizon_state: HorizonState of the target
            observer: Position of the observer
        """
        ...
